
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include <iostream.h>
#include <iomanip.h>

#include "base/ObjId.h"
#include "base/RndPermut.h"
#include "base/polyLogCats.h"
#include "base/polyLogTags.h"
#include "client/IcpCltXact.h"
#include "runtime/HostMap.h"
#include "runtime/StatPhase.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/LogComment.h"
#include "runtime/ErrorMgr.h"
#include "runtime/EphPortMgr.h"
#include "runtime/ExpPortMgr.h"
#include "runtime/polyBcastChannels.h"
#include "runtime/polyErrors.h"
#include "runtime/globals.h"
#include "csm/ContentSel.h"
#include "csm/ContentCfg.h"
#include "pgl/RobotSym.h"
#include "client/CltConnMgr.h"
#include "client/CltOpts.h"
#include "client/CltXact.h"
#include "client/PopModel.h"
#include "client/ServerRep.h"
#include "client/PrivCache.h"
#include "client/CltCfg.h"
#include "client/Client.h"


class IcpAgentCfg;

struct SharedCfgItem { const RobotSym *sym; CltCfg *cfg; };
static Array<SharedCfgItem> SharedCfgs;

XactFarm<CltXact> *Client::TheXacts = 0;
ObjFarm<IcpCltXact> Client::TheIcpXacts;

Array<PortMgr*> Client::ThePortMgrs;
Array<ServerRep*> Client::TheServers;
Ring<int> *Client::TheWarmupPlan = 0;

int Client::TheClientCnt = 0;


Client::Client(): thePrivCache(0), theCfg(0), theConnMgr(0), 
	theIcpClient(0),
	theCcXactLvl(0), theLaunchCount(0), theExtraLaunchLvl(0) {

	theChannels.append(TheInfoChannel);
	theChannels.append(TheLogCfgChannel);
	theChannels.append(TheLogStateChannel);
	startListen();
}

Client::~Client() {
	delete thePrivCache;
}

void Client::configure(const RobotSym *cfg, const NetAddr &aHost, bool bind2iface) {
	Assert(TheXacts);

	SockOpt opt;
	Agent::configure(cfg, aHost, opt);
	theHost.port(-1); // clients do not care about port number

	theCfg = sharedConfig(cfg);

	thePrivWorlds.stretch(theCfg->theSrvCount);
	for (int i = 0; i < theCfg->theSrvCount; ++i)
		thePrivWorlds.append(PrivWorld());

	int pcCap = 0;
	if (cfg->privCache(pcCap))
		thePrivCache = new PrivCache(pcCap);

	int openConnLmt = -1;
	cfg->openConnLimit(openConnLmt);
	theConnMgr = new CltConnMgr;
	theConnMgr->configure(opt, cfg->pconnUseLmt(), openConnLmt,
		TheCltOpts.thePxyAddr ? 1 : theCfg->theSrvCount);
	theConnMgr->portMgr(getPortMgr(bind2iface));
	theConnMgr->idleTimeout(cfg->idlePconnTimeout());

	if (theCfg->thePeerHttp && !theCfg->thePeerIcp) {
		cerr << cfg->loc() << "the HTTP peer is at " << theCfg->thePeerHttp << ", but where is the ICP peer?" << endl;
		exit(-3);
	}

	TheXacts->limit(1024); // magic, no good way to estimate
}

void Client::icpClient(IcpClient *anIcpClient) {
	Assert(!theIcpClient);
	theIcpClient = anIcpClient;
}

void Client::start() {
	Assert(!theLaunchCount);
	Assert(theConnMgr);

	Agent::start();

	// first client that gets here will lock the phase and build the plan
	if (!TheClientCnt) {
		TheStatPhaseMgr->lock();
		BuildWarmupPlan();
	}

	TheClientCnt++;

	if (theCfg->theLaunchWin > 0) {
		// uniformly distributed start delay
		static RndGen rng;
		const double delay = rng(0.0, theCfg->theLaunchWin.secd());
		sleepFor(Time::Secd(delay));
	} else {
		scheduleLaunch(TheClock);
	}
}

void Client::noteInfoEvent(BcastChannel *ch, InfoEvent ev) {
	Assert(ch == TheInfoChannel);
	if (ev == ieWorkSetFreeze) {
		for (int i = 0; i < thePrivWorlds.count(); ++i)
			thePrivWorlds[i].freezeWss();

		static bool frozen = false;
		if (!frozen) {
			FreezeSrvWss();
			frozen = true;
		}
	}
}

void Client::noteLogEvent(BcastChannel *ch, OLog &log) {
	if (ch == TheLogCfgChannel) {
		log << bege(lgCltCfg, lgcCltSide);
		// XXX: implement (CltCfgRec)
		log << ende;
	} else
	if (ch == TheLogStateChannel) {
		log << bege(lgCltState, lgcCltSide)
			<< theSeqvId 
			<< theLaunchCount
			<< thePrivWorlds
			<< ende;
	}
}

void Client::noteXactDone(CltXact *x) {
	Assert(x);

	theConnMgr->put(x->conn());
	TheXacts->put(x);
	theCcXactLvl--;
	Assert(theCcXactLvl >= 0);

	// push waiting xactions forward
	if (theLaunchDebts.count() && !theConnMgr->atHardConnLimit())
		resumeXact();
}

void Client::scheduleLaunch(Time lastLaunch) {
	Assert(!thePendAlarmCnt);
	if (theThinkDistr) {
		const double lf = TheStatPhaseMgr->loadFactor(lastLaunch, theLaunchCount);
		Assert(lf >= 0);

		if (lf <= 0) // zero load: no more requests needed
			return;

		double delay;
		if (TheStatPhaseMgr->varyLoad()) {
			// next launch should happen [0, 2dt] from last launch
			// where dt is 1/current_rate
			static RndGen rng;
			delay = rng(0.0, 2.*theThinkDistr->mean()/lf);
		} else {
			delay = theThinkDistr->trial()/lf;
		}
		//cerr << here << TheClock << "| " << seqId() << " delay: " << delay << " lf: " << lf << endl;
		sleepTill(lastLaunch + Time::Secd(delay));
	} else {
		tryLaunch(genXact());
	}
}

// will create a xaction if needed
// must not be called directly but rather through tryLaunch
bool Client::launch(CltXact *x) {
	Assert(x);
	Assert(!theConnMgr->atHardConnLimit());

	const NetAddr &connAddr = nextHopAddr(x);
	Assert(connAddr);

	if (Connection *conn = theConnMgr->get(connAddr)) {
		theLaunchCount++;
		theCcXactLvl++;
		x->exec(this, conn);
		return true;
	}

	if (ReportError(errno))
		Comment(1) << theHost << " failed to connect to " << connAddr << endc;
	Broadcast(TheXactErrChannel, x);

	TheXacts->put(x);

	// if no alarms are pending, we may get stuck and submit
	// no more xactions (e.g. best effort); this is a kludge
	if (!thePendAlarmCnt && !theCcXactLvl)
		sleepTill(TheClock); // submit more next cycle

	return false;
}

// will launch the xaction (if possible) or postpone it (if not)
// returns true if the xaction will be launched [eventually]
bool Client::tryLaunch(CltXact *x) {
	Assert(x);

	// determine request type
	x->reqType((int)theCfg->theRTypeSel->trial());

	// should we ask peers for the best server?
	if (theCfg->thePeerIcp && !x->askedPeer()) {
		askPeer(theCfg->thePeerIcp, x);
		return true;
	}

	// check if we should postpone the xaction
	if (theConnMgr->atHardConnLimit()) {
		if (theCfg->theWaitXactLmt < 0 || theLaunchDebts.count() < theCfg->theWaitXactLmt) {
			suspendXact(x);
			return true;
		}

		if (ReportError(errTooManyWaitXact))
			Comment(3) << "xactions active: " << theCcXactLvl 
				<< " waiting: " << theLaunchDebts.count() 
				<< " limit: " << theCfg->theWaitXactLmt << endc;
		TheXacts->put(x);
		return false;
	}

	return launch(x);
}

void Client::suspendXact(CltXact *x) {
	theLaunchDebts.append(x);
	Broadcast(TheWaitBegChannel, x);
}

void Client::resumeXact() {
	CltXact *x = theLaunchDebts.pop();
	Broadcast(TheWaitEndChannel, x);
	launch(x);
}

CltXact *Client::genXact(const OidRec &or) {
	CltXact *x = TheXacts->get();
	x->oidRec(or);
	return x;
}

CltXact *Client::genXact() {
	OidRec or;

	// at the beginning, touch all servers so we have
	// private or secured id for each server to generate hits
	if (TheWarmupPlan)
		or.sidx = genWarmupSrvIdx();

	if (or.sidx < 0)
		or.sidx = (int)theCfg->theOriginSel->trial();

	or.hidx = TheServers[or.sidx]->hostIdx();

	genOid(or);
	return genXact(or);
}

void Client::genOid2(OidRec &or) {
	ServerRep &srvRep = *GetServerRep(or.sidx);
	PrivWorld &privWorld = thePrivWorlds[or.sidx];
	//cerr << here << "pid: " << theId << " server: " << or.sidx << " wid: " << srvRep.id() << " sec oid: " << srvRep.secOid() << endl;

	const bool forcePubOid = srvRep.secOid() <= 0 && srvRep.haveNewOids();

	static RndGen rng;
	const bool genPubOid = forcePubOid || rng.event(theCfg->thePublicRatio);
	const bool repeatOid = rng.event(theCfg->theRecurRatio*TheStatPhaseMgr->recurFactor());

	/* this logic gives priority to repeatOid goal rather than genPubOid */

	if (repeatOid) {
		or.repeat = true;

		const int privOid = privWorld.lastObjName();
		const int privWss = privWorld.wss();
		const bool privCanRep = privOid && privWss;

		const int pubOid = srvRep.secOid();
		const int pubWss = srvRep.wss();
		const bool pubCanRep = pubOid && pubWss;

		if (pubCanRep && (genPubOid || !privCanRep)) {
			or.wid = srvRep.id(); // use server id
			or.oid = theCfg->thePopModel->choose(rng, 
				pubOid, pubWss, srvRep.hotSetPos());
			return;
		}

		if (privCanRep) {
			or.wid = theId; // use this robot id
			or.oid = theCfg->thePopModel->choose(rng,
				privOid, privWss, privWorld.hotSetPos());
			return;
		}

		// failed to repeat, fall through
	}

	or.repeat = false;
	if (genPubOid) {
		// can ask for a new oid?
		if (srvRep.haveNewOids()) {
			or.wid = srvRep.id(); // use server id
			or.oid = srvRep.newOid();
			return;
		}

		// report insufficient number of new oids if server is initialized
		if (srvRep.secOid() > srvRep.maxNewOidCap()) {
			ReportError(errNoNewOids);
			srvRep.noteNoNewOids();
		}
		// failed to produce new public oid, fall through
	}
		
	// new private object [last resort, never fails]
	or.wid = theId; // use this robot id
	or.oid = privWorld.nextObjName();
}

void Client::genOid(OidRec &or) {
	genOid2(or);
	if (theCfg->genUniqUrls)
		or.wid = UniqId::Create(); // changes every time
	const HostCfg *hcfg = TheHostMap->at(or.hidx);
	Assert(hcfg);
	Assert(hcfg->theContent);
	const ContentCfg *ccfg = hcfg->theContent->getDir(or);
	or.tid = ccfg->id();
}

void Client::noteEmbedded(CltXact *, const OidRec &or) {
	static RndGen rng;
	if (!rng.event(theCfg->theEmbedRecurRatio))
		return;

	if (tryLaunch(genXact(or)))
		theExtraLaunchLvl++;
}

void Client::askPeer(const NetAddr &addr, CltXact *x) {
	IcpCltXact *q = TheIcpXacts.get();
	q->reason(x);
	q->exec(this, addr);
}

void Client::notePeerAsked(IcpCltXact *q) {
	Assert(q);

	CltXact *x = q->reason(this);
	x->usePeer(q->hit());
	/* pass ICP stats here ... */

	TheIcpXacts.put(q);

	tryLaunch(x);
}

// select the address for the next hop connection
const NetAddr &Client::nextHopAddr(const CltXact *x) const {

	if (x->usePeer() && theCfg->thePeerHttp)
		return theCfg->thePeerHttp;

	if (TheCltOpts.thePxyAddr)
		return TheCltOpts.thePxyAddr;

	return GetServerRep(x->oidRec().sidx)->addr();
}

// prepare server representatives (used by sharedConfig())
void Client::scanServers(const RobotSym *rs) {
	Array<NetAddr*> addrs;
	Assert(rs->origins(addrs));

	const int savedCount = TheServers.count();
	for (int i = 0; i < addrs.count(); ++i)
		cAddServer(*addrs[i], !savedCount);

	if (savedCount && TheServers.count() != savedCount) {
		Comment(1) << "varying the number of origin servers per robot "
			<< "is not yet supported for memory consumption reasons; "
			<< "let us know if you need this feature" << endl;
		FatalError(errOther);
	}

	// get rid of addresses
	while (addrs.count())
		delete addrs.pop();
}

PortMgr *Client::getPortMgr(bool bind2iface) {
	const NetAddr addr = bind2iface ? theHost : NetAddr();

	// check if we already have a port mgr for host
	for (int i = 0; i < ThePortMgrs.count(); ++i) {
		if (ThePortMgrs[i]->addr() == addr)
			return ThePortMgrs[i];
	}

	// create new port manager
	PortMgr *mgr = TheCltOpts.thePorts.set() ?
		(PortMgr*) new ExpPortMgr(addr, TheCltOpts.thePorts.lo(), TheCltOpts.thePorts.hi()) :
		(PortMgr*) new EphPortMgr(addr);
	ThePortMgrs.append(mgr);
	return mgr;
}

bool Client::srvReadyToHit(const ServerRep &s) const {
	// already got secured oid OR
	// was accessed and will never get a secured oid
	return s.secOid() > 0 || (theCfg->genUniqUrls && s.accesses() > 0);
}

int Client::genWarmupSrvIdx() {
	Assert(TheWarmupPlan);
	int idx = -1;

	static int touchCnt = 0;
	while (!TheWarmupPlan->empty() && idx < 0) {
		const int i = TheWarmupPlan->dequeue();

		if (srvReadyToHit(*TheServers[i]))
			continue;

		TheWarmupPlan->enqueue(i);
		idx = i;
		touchCnt++;
	}

	if (TheWarmupPlan->count()) {
		// once in a while, report progress
		const Time reportGap = Time::Sec(45); // how often to report
		static const Time start = TheClock;
		static Time nextReport = start + reportGap;
		Clock::Update(false);
		if (nextReport <= TheClock) {

			int readyCnt = 0;
			for (int i = 0; i < TheServers.count(); ++i) {
				if (srvReadyToHit(*TheServers[i]))
					readyCnt++;
			}

			const double readyp = Percent(readyCnt, TheServers.count());
			const double touchp = touchCnt >= 2*TheServers.count() ?
				readyp : Percent(touchCnt, 2*TheServers.count());
			const double progrp = MiniMax(touchp, readyp, 99.0);
			Comment(5) << "fyi: server scan is probably "
				<< setw(6) << progrp << "% completed with " 
				<< setw(5) << readyCnt << " out of "
				<< TheServers.count() << " servers ("
				<< setw(6) << readyp
				<< "%) ready to be hit" << endc;
			nextReport = TheClock + reportGap;
		}
	} else {
		delete TheWarmupPlan;
		TheWarmupPlan = 0;

		Comment(3) << "fyi: server scan completed with all " << TheClientCnt << " local robots "
			<< "ready to hit all " << TheServers.count() << " servers" << endc;
		TheStatPhaseMgr->unlock();
	}
	
	return idx;
}

// creates new server representative if needed
// returns true iff creation was needed
bool Client::cAddServer(const NetAddr &addr, bool forceAdd) {
	if (!forceAdd) {
		// check if we already know about this server
		for (int i = 0; i < TheServers.count(); ++i) {
			if (TheServers[i]->addr() == addr)
				return false;
		}
	}
	
	// find host index for the new server
	Assert(TheHostMap);
	int idx = -1;
	if (!TheHostMap->find(addr, idx)) {
		Comment << here << "Robot@" << theHost
			<< " cannot find configuration for server@" << addr
			<< endc;
		FatalError(errOther);
	}

	TheServers.append(new ServerRep(addr, idx));
	return true;
}

ServerRep *Client::GetServerRep(int idx) {
	Assert(0 <= idx && idx < TheServers.count());
	return TheServers[idx];
}


CltCfg *Client::sharedConfig(const RobotSym *rs) {
	CltCfg *cfg =0;
	for (int i = 0; i < SharedCfgs.count(); ++i) {
		if (SharedCfgs[i].sym == rs) {
			cfg = SharedCfgs[i].cfg;
			break;
		}
	}

	if (!cfg) {
		cfg = new CltCfg;
		cfg->configure(rs);
		SharedCfgItem ci;
		ci.sym = rs;
		ci.cfg = cfg;
		SharedCfgs.append(ci);
		// do this once per shared configuration
		scanServers(rs);
	}

	return cfg;
}

void Client::FreezeSrvWss() {
	for (int i = 0; i < TheServers.count(); ++i)
		TheServers[i]->freezeWss();
	Comment(5) << "fyi: public wss will be frozen"
		<< " with at least " << ServerRep::TheSharedWss
		<< " objects total" << endc;
}

void Client::BuildWarmupPlan() {
	Assert(!TheWarmupPlan);

	TheWarmupPlan = new Ring<int>(TheServers.count());

	// populate with server indeces
	for (int idx = 0; idx < TheServers.count(); ++idx)
		TheWarmupPlan->enqueue(idx);

	TheWarmupPlan->randomize();
}

static
OLog &operator <<(OLog &log, ServerRep *sr) {
	sr->store(log);
	return log;
}

void Client::Farm(XactFarm<CltXact> *aFarm) {
	Assert(!TheXacts && aFarm);
	TheXacts = aFarm;
}

void Client::LogState(OLog &log) {
	log << bege(lgSrvRepState, lgcCltSide) << TheServers << ende;
}
