
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "polygraph.h"

#include "xstd/gadgets.h"
#include "base/OLog.h"
#include "base/ILog.h"
#include "runtime/httpHdrs.h"
#include "runtime/ErrorMgr.h"
#include "runtime/polyErrors.h"
#include "client/ServerRep.h"


int ServerRep::TheSharedWss = -1;
int ServerRep::TheNewOidsPrefetchDepth = -1;
int ServerRep::TheNewOidsPerReqMax = -1;

void ServerRep::ConfigureNids(int prefetchLmt, int perReqMax) {
	TheNewOidsPrefetchDepth = prefetchLmt;
	TheNewOidsPerReqMax = perReqMax;
}

ServerRep::ServerRep(const NetAddr &anAddr, int aHostIdx): 
	theAddr(anAddr), theHostIdx(aHostIdx),
	theLastNewOid(0), theSecOid(0), 
	theAccCount(0), theWss(-1), thePendCnt(0) {
	if (TheNewOidsPrefetchDepth > -1) {
		theNewOids.resize(TheNewOidsPrefetchDepth);
		theFailedOids.resize(TheNewOidsPrefetchDepth);
	}
}

int ServerRep::oidShortage() {
	//cerr << here << TheNewOidsPerReqMax << ": " << theNewOids.capacity() << " - (" << thePendCnt << " + " << theNewOids.count() << ") = " << shortage << " : " << req_size << endl;
	const int shortage = theNewOids.capacity() - (thePendCnt + theNewOids.count());
	const int req_size = Min(TheNewOidsPerReqMax, shortage);
	thePendCnt += req_size;
	return req_size;
}

int ServerRep::newOid() {
	Assert(haveNewOids());
	return theLastNewOid = theNewOids.dequeue();
}

int ServerRep::failedOid() {
	Assert(haveFailedOids());
	return theFailedOids.dequeue();
}

void ServerRep::noteNewOids(int oid, int askCnt, int gotCnt) {
	Assert(!gotCnt || oid > 0);

	thePendCnt -= askCnt;
	if (thePendCnt < 0)
		thePendCnt = 0;

	while (gotCnt-- > 0) {
		if (theNewOids.full()) {
			ReportError(errNidRepFlood);
			break;
		}
		theNewOids.enqueue(oid++);
	}
}

void ServerRep::noteFailedOid(int oid) {
	Assert(oid > 0);
	if (oid <= theSecOid) // no need to report
		return;

	if (theFailedOids.full())
		ReportError(errFidFlood);
	else
		theFailedOids.enqueue(oid);
}

void ServerRep::syncSrvInfo(const RepHdr &rep) {
	syncSrvId(rep.theSrvWid);

	if (rep.theSecOid > theSecOid)
		theSecOid = rep.theSecOid;

	syncWss(rep.theSrvWss);

	theHotSetPos = rep.theHotSetPos;
}

void ServerRep::noteNoNewOids() {
	if (thePendCnt + theNewOids.count() >= theNewOids.capacity())
		theNewOids.resize(Max(1, 2*theNewOids.capacity()));
}

void ServerRep::freezeWss() {
	// do not allow temporary lockup with theSecOid to affect WS
	syncWss(Max(theSecOid, theLastNewOid));
}

void ServerRep::syncSrvId(const UniqId &aSrvId) {
	if (aSrvId) {
		if (!theId)
			theId = aSrvId;
		else
		if (theId != aSrvId) {
			ReportError(errSrvChangedWid);
			theNewOids.reset();
			theId = aSrvId;
			theSecOid = 0;
			thePendCnt = 0;
		}
	}
}

void ServerRep::syncWss(int newWss) {
	if (newWss > theWss) {
		if (TheSharedWss < 0)
			TheSharedWss = newWss;
		else
		if (theWss >= 0)
			TheSharedWss -= theWss;
		TheSharedWss += newWss;

		theWss = newWss;
	}
}

void ServerRep::store(OLog &log) const {
	log
		<< theAddr
		<< theId
		<< theSecOid
		<< theWss
		;
}

void ServerRep::load(ILog &log) {
	log
		>> theAddr
		>> theId
		>> theSecOid
		>> theWss
		;
}
