
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl/pgl.h"

#include "xstd/String.h"
#include "xstd/TblDistr.h"
#include "xstd/rndDistrs.h"
#include "pgl/PglBoolSym.h"
#include "pgl/PglIntSym.h"
#include "pgl/PglNumSym.h"
#include "pgl/PglRateSym.h"
#include "pgl/PglRec.h"
#include "pgl/PglStringSym.h"
#include "pgl/PglArraySym.h"
#include "pgl/PglNetAddrSym.h"
#include "pgl/PopModelSym.h"
#include "pgl/RobotSym.h"


RobotSym::RobotSym(): AgentSym("Robot") {
	theRec->bAdd("addr[]", "origins", 0);
	theRec->bAdd("time", "launch_win", 0);
	theRec->bAdd("rate", "req_rate", 0);
	theRec->bAdd("time_distr", "req_inter_arrival", 0);
	theRec->bAdd("float", "public_interest", 0);
	theRec->bAdd("float", "recurrence", 0);
	theRec->bAdd("float", "embed_recur", 0);
	theRec->bAdd("PopModel", "pop_model", 0);
	theRec->bAdd("string[]", "req_types", 0);
	theRec->bAdd("int", "private_cache_cap", 0);
	theRec->bAdd("bool", "unique_urls", 0);
	theRec->bAdd("int", "open_conn_lmt", 0);
	theRec->bAdd("int", "wait_xact_lmt", 0);
	theRec->bAdd("int", "icp_port", 0);
	theRec->bAdd("addr", "peer_http", 0);
	theRec->bAdd("addr", "peer_icp", 0);
}

RobotSym::RobotSym(const String &aType, PglRec *aRec): AgentSym(aType, aRec) {
}

bool RobotSym::isA(const String &type) const {
	return AgentSym::isA(type) || type == "Robot";
}

SynSym *RobotSym::clone(const String &type) const {
	if (isA(type))
		return new RobotSym(this->type(), theRec->clone());
	return AgentSym::clone(type);
}

bool RobotSym::reqRate(double &rate) const {
	SynSymTblItem *rri = 0;
	Assert(theRec->find("req_rate", rri));
	if (rri->sym())
		rate = ((const RateSym&)rri->sym()->cast("rate")).val();
	return rri->sym();
}

RndDistr *RobotSym::reqInterArrival() const {
	RndDistr *iad = getDistr("req_inter_arrival");
	double rr = -1;
	const bool hasRR = reqRate(rr);
	if (hasRR && iad)
		cerr << loc() << ": cannot specify both request rate"
			<< " and inter-arrival distribution" << endl << xexit;

	if (iad)
		return iad;

	if (!hasRR)
		return 0;

	if (rr <= 0)
		cerr << loc() << ": request rate for robot " << kind() << " is not positive" << endl << xexit;

	// poisson request stream (default) is modeled using an exponential
	// distribution with a mean of a given request rate
	static RndGen rng; // all clients should share one rng
	return new ExpDistr(&rng, 1/rr);
}

Time RobotSym::launchWin() const {
	return getTime("launch_win");
}

bool RobotSym::publicRatio(double &ratio) const {
	return getDouble("public_interest", ratio);
}

bool RobotSym::recurRatio(double &ratio) const {
	return getDouble("recurrence", ratio);
}

bool RobotSym::embedRecurRatio(double &ratio) const {
	return getDouble("embed_recur", ratio);
}

PopModelSym *RobotSym::popModel() const {
	SynSymTblItem *pmi = 0;
	Assert(theRec->find("pop_model", pmi));
	return pmi->sym() ?
		&((PopModelSym&)pmi->sym()->cast("PopModel")) : 0;
}

bool RobotSym::uniqueUrls(bool &set) const {
	return getBool("unique_urls", set);
}

bool RobotSym::openConnLimit(int &lmt) const {
	return getInt("open_conn_lmt", lmt);
}

bool RobotSym::waitXactLimit(int &lmt) const {
	return getInt("wait_xact_lmt", lmt);
}

bool RobotSym::icpPort(int &port) const {
	return getInt("icp_port", port);
}

NetAddr RobotSym::peerHttp() const {
	return getNetAddr("peer_http");
}

NetAddr RobotSym::peerIcp() const {
	return getNetAddr("peer_icp");
}

bool RobotSym::privCache(int &capacity) const {
	return getInt("private_cache_cap", capacity);
}

bool RobotSym::origins(Array<NetAddr*> &addrs) const {
	SynSymTblItem *oi = 0;
	Assert(theRec->find("origins", oi));
	if (!oi->sym())
		return false; // undefined

	// build address array and selector
	ArraySym &os = (ArraySym&)oi->sym()->cast("Array");
	for (int i = 0; i < os.count(); ++i) {
		const NetAddr &addr = ((const NetAddrSym&)os[i]->cast("addr")).val();
		addrs.append(new NetAddr(addr));
	}

	return addrs.count();
}

bool RobotSym::origins(Array<NetAddr*> &addrs, RndDistr *&selector) const {
	SynSymTblItem *oi = 0;
	Assert(theRec->find("origins", oi));
	if (!oi->sym())
		return false; // undefined

	// build address array and selector
	origins(addrs);
	ArraySym &os = (ArraySym&)oi->sym()->cast("Array");
	selector = os.makeSelector(kind() + "-origins");
	return true;
}

RndDistr *RobotSym::reqTypes(const Array<String*> &typeNames) const {
	Assert(typeNames.count());

	SynSymTblItem *rti = 0;
	Assert(theRec->find("req_types", rti));
	if (!rti->sym())
		return 0; // undefined

	// build selector using name index
	ArraySym &rts = (ArraySym&)rti->sym()->cast("Array");
	Array<double> probs_in(typeNames.count());
	Array<double> probs_out(typeNames.count());
	rts.copyProbs(probs_in);

	for (int i = 0; i < probs_in.count(); ++i) {
		const String &typeName = ((const StringSym&)rts[i]->cast("string")).val();

		int idx = -1;
		for (int t = 0; idx < 0 && t < typeNames.count(); ++t) {
			if (*typeNames[t] == typeName)
				idx = t;
		}
			
		if (idx < 0) {
			cerr << loc() << "unknown request type `" << typeName
				<< "'; known types are:";
			for (int n = 0; n < typeNames.count(); ++n)
				cerr << " \"" << *typeNames[n] << '"';
			cerr << endl;
			exit(-2);
		}

		if (probs_in[i] > 0)
			probs_out.put(probs_in[i], idx);
	}

	if (!probs_out.count())
		return 0;

	return TblDistr::FromDistrTable(kind() + "-req_types", probs_out);
}
