
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include <math.h>
#include <iomanip.h>

#include "xstd/Clock.h"
#include "base/Progress.h"
#include "base/polyLogCats.h"
#include "base/polyLogTags.h"
#include "base/ErrorStrTbl.h"
#include "pgl/StatsSampleCfg.h"
#include "pgl/StatsSampleSym.h"
#include "pgl/GoalSym.h"
#include "pgl/PhaseSym.h"
#include "runtime/Connection.h"
#include "runtime/Goal.h"
#include "runtime/Xaction.h"
#include "runtime/IcpXaction.h"
#include "runtime/LogComment.h"
#include "runtime/PolyOLog.h"
#include "runtime/StatsSampleMgr.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/StatPhase.h"
#include "runtime/globals.h"
#include "runtime/polyErrors.h"
#include "runtime/polyBcastChannels.h"


StatPhase::StatPhase(): theRecs(lgcEnd), theMgr(0),
	wasUsed(false), logStats(true), wasStopped(false),
	theLoadFactBeg(1), theLoadFactEnd(1), theLoadFactK(0),
	theRecurFact(1),
	theLockCount(0) {

	for (int i = 0; i < theRecs.capacity(); ++i)
		theRecs.push(new StatPhaseRec);

	theChannels.append(TheErrChannel);
}

StatPhase::~StatPhase() {
	while (theRecs.count()) delete theRecs.pop();
}

void StatPhase::configure(const PhaseSym *cfg, const StatPhase *prevPh) {
	Assert(cfg);

	name(cfg->name());
	theGoal.configure(cfg->goal());
	for (int i = 0; i < theRecs.count(); ++i)
			theRecs[i]->theGoal = theGoal;

	cfg->logStats(logStats);

	cfg->recurFactor(theRecurFact);

	// use loadFactorEnd from the previous phase if needed
	if (!cfg->loadFactorBeg(theLoadFactBeg) && prevPh)
		theLoadFactBeg = prevPh->theLoadFactEnd;

	// preserve load level unless instructed otherwise
	if (!cfg->loadFactorEnd(theLoadFactEnd))
		theLoadFactEnd = theLoadFactBeg;

	configureSamples(cfg);

	const double fdiff = theLoadFactEnd - theLoadFactBeg;

	// varying load is possible only if goal is specified
	if (fdiff && !theGoal) {
		cerr << cfg->loc() << "phase named `" << theName
			<< "' has load factors but has no positive goal" << endl;
		exit(-2);
	}

	// cache commonly used value
	// note: duration takes priority over xact count if both are given
	theLoadFactK =
		theGoal.duration() > 0 ?
			fdiff/theGoal.duration().secd() :
		(theGoal.xactCnt() > 0 ?
			fdiff/theGoal.xactCnt() : 0);

	theErrs.maxErrNo(polyErrorStrTbl()->maxErrNo());
}

void StatPhase::configureSamples(const PhaseSym *cfg) {
	Array<const StatsSampleSym*> ssyms;
	if (cfg->statsSamples(ssyms)) {
		for (int i = 0; i < ssyms.count(); ++i) {
			const StatsSampleSym &ssym = *ssyms[i];

			StatsSampleCfg cfg;
			// XXX: append counter to phase names
			cfg.name = ssym.name() ? ssym.name() : name();
			Assert(ssym.capacity(cfg.capacity));
			cfg.start = ssym.start();

			if (cfg.start >= 0) {
				TheStatsSampleMgr.addSample(cfg);
			} else {
				theSamples.append(new StatsSampleCfg(cfg));
				TheStatsSampleMgr.willAddSample();
			}
		}
	}
	// no support for multiple pending samples for now
	Assert(theSamples.count() <= 1); 
}

double StatPhase::loadFactorBeg() const {
	return theLoadFactBeg;
}

double StatPhase::loadFactorEnd() const {
	return theLoadFactEnd;
}

double StatPhase::loadFactor(Time lastFactTm, int lastFactCnt) const {
	const double t = theGoal.duration() > 0 ?
		Max((lastFactTm - theIntvlStart).secd(), 0.0) : (double)lastFactCnt;
	const double lastLF = theLoadFactK*t + theLoadFactBeg;
	const double d = lastLF*lastLF + 4*theLoadFactK;
	const double newLF = d >= 0 ? (sqrt(d) + lastLF)/2 : (lastLF/2);

	// make sure load factor stays between load_fact_beg and load_fact_end
	return theLoadFactBeg <= theLoadFactEnd ?
		MiniMax(theLoadFactBeg, newLF, theLoadFactEnd):
		MiniMax(theLoadFactEnd, newLF, theLoadFactBeg);
}

double StatPhase::recurFactor() const {
	return theRecurFact;
}

void StatPhase::lock() {
	theLockCount++;
}

void StatPhase::unlock() {
	if (locked()) {
		theLockCount--;
		if (!locked())
			checkpoint();
	} else {
		Comment(0) << "internal error: attempt to unlock an unlocked phase (salvaged)" << endl;
	}
}

bool StatPhase::locked() const {
	return theLockCount > 0;
}

Time StatPhase::duration() const {
	return TheClock.time() >= theIntvlStart ?
		TheClock - theIntvlStart : Time(0,0);
}

int StatPhase::xactCnt() const {
	int count = 0;
	for (int i = 0; i < theRecs.count(); ++i)
		count += theRecs[i]->complCnt();
	return count;
}

BigSize StatPhase::fillSz() const {
	BigSize fill(0);
	for (int i = 0; i < theRecs.count(); ++i)
		fill += theRecs[i]->totFillSize();
	return fill;
}

int StatPhase::errCnt() const {
	int count = 0;
	for (int i = 0; i < theRecs.count(); ++i)
		count += theRecs[i]->theXactErrCnt;
	return count;
}

double StatPhase::errRatio() const {
	int xcount = 0, ecount = 0;
	for (int i = 0; i < theRecs.count(); ++i) {
		xcount += theRecs[i]->complCnt();
		ecount += theRecs[i]->theXactErrCnt;
	}
	return Ratio(ecount, xcount + ecount);
}


void StatPhase::start(StatPhaseMgr *aMgr, const StatPhase *prevPhase) {
	Assert(!theMgr);
	theMgr = aMgr;

	if (prevPhase) {
		for (int i = 0; i < theRecs.count(); ++i)
			theRecs[i]->keepLevels(*prevPhase->theRecs[i]);
	}

	TheOLog << bege(lgStatPhaseBeg, lgcAll) << name() << ende;

	startListen();
	theIntvlStart = TheClock;

	for (int i = 0; i < theSamples.count(); ++i) {
		StatsSampleCfg &cfg = *theSamples[i];
		if (cfg.start < 0) {
			cfg.start = theIntvlStart;
			if (theGoal.duration() > 0)
				cfg.start += theGoal.duration()/2; // middle of the phase
			cfg.start -= Clock::TheStartTime;
		}
		TheStatsSampleMgr.addSample(cfg);
	}

	// manager should be careful with immediate "returns"
	if (!checkpoint()) { 
		// not an immediate exit
		wasUsed = true; 
		if (theGoal.duration() > 0)
			sleepFor(theGoal.duration());
	}
}

void StatPhase::noteConnEvent(BcastChannel *ch, const Connection *c) {
	StatIntvl::noteConnEvent(ch, c);
	if (ch == TheConnCloseChannel && !c->bad()) {
		StatPhaseRec &rec = *theRecs[c->logCat()];
		const Time ttl = TheClock - c->openTime();
		rec.theConnClose.record(c->closeKind(), ttl, c->useCnt());
		rec.theConnLifeVsUse.record(ttl, c->useCnt());
	}
	checkpoint();
}

void StatPhase::noteXactEvent(BcastChannel *ch, const Xaction *x) {
	StatIntvl::noteXactEvent(ch, x);

	if (ch == TheXactEndChannel) {
		StatPhaseRec &rec = *theRecs[x->logCat()];
		rec.theHttpXacts.record(x->lifeTime(), x->repSize(), x->oidRec().hit);
		rec.theRptmVsSz.record(x->lifeTime(), x->repSize());
		rec.theContType.record(x->oidRec().tid, x->repSize());
	}
	checkpoint();
}

void StatPhase::noteIcpXactEvent(BcastChannel *ch, const IcpXaction *x) {
	StatIntvl::noteIcpXactEvent(ch, x);

	if (ch == TheIcpXactEndChannel) {
		StatPhaseRec &rec = *theRecs[x->logCat()];
		if (!x->timedout())
			rec.theIcpXacts.record(x->lifeTime(), x->repSize(), x->hit());
	} else {
		Assert(false);
	}
	checkpoint();
}

void StatPhase::noteErrNoEvent(BcastChannel *ch, int errNo) {
	Assert(ch == TheErrChannel);
	theErrs.record(errNo);
}

void StatPhase::name(const String &aName) {
	theName = aName;
	if (!theName)
		theName = "";
	for (int i = 0; i < theRecs.count(); ++i)
		theRecs[i]->theName = theName;
}

void StatPhase::statsLogged(bool are) {
	logStats = are;
}

void StatPhase::wakeUp(const Alarm &alarm) {
	StatIntvl::wakeUp(alarm);
	Assert(locked() || checkpoint());
}

void StatPhase::flush() {
	stop();

	if (logStats) {
		TheOLog << bege(lgErrorStat, lgcAll) << (Array<int>&)theErrs << ende;
		storeAll(TheOLog, lgStatPhaseRec);
	}
}

bool StatPhase::checkpoint()  {
	if (!locked() && theGoal.reached(*this)) {
		stop();
		report();

		Assert(theMgr);
		StatPhaseMgr *mgr = theMgr;
		theMgr = 0;
		mgr->noteDone(this);
		return true;
	}
	return false;
}

void StatPhase::stop() {
	Assert(theIntvlStart >= 0);
	if (wasStopped) // already stopped
		return;
	wasStopped = true;

	setDuration(theIntvlStart);
	stopListen();
	TheAlarmClock.cancelAll(this);

	TheOLog << bege(lgStatPhaseEnd, lgcAll) << name() << ende;
}

StatIntvlRec &StatPhase::getRec(int cat) {
	Assert(0 < cat && cat < theRecs.count());
	return *theRecs[cat];
}

const StatIntvlRec &StatPhase::getRec(int cat) const {
	Assert(0 < cat && cat < theRecs.count());
	return *theRecs[cat];
}

// XXX: we probably should report _all_ active sides
void StatPhase::report() const {
	Comment(1) << "p-" << name();
	Comment << ' ' << setw(6) << TheProgress.xacts();
	if (TheReportCat > 0)
		theRecs[TheReportCat]->linePrint(Comment);
	Comment << endc;
}


void StatPhase::reportCfg(ostream &os) const {
	os 
		<< setw(10) << name()
		<< ' ' << setw(8) << theLoadFactBeg 
		<< ' ' << setw(8) << theLoadFactEnd
		<< ' ' << setw(5) << theGoal
		<< "\t " << (logStats ? "" : "!log")
		<< endl;
}
