
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include "base/OLog.h"
#include "xstd/String.h"
#include "xstd/Clock.h"
#include "xstd/gadgets.h"
#include "runtime/StrIdentifier.h"
#include "runtime/HttpDate.h"
#include "runtime/httpHdrs.h"


/* internal type to store static parsing info */
class MsgHdrParsTab {
	public:
		MsgHdrParsTab();
		~MsgHdrParsTab();

	public:
		StrIdentifier *ids;
		Array<MsgHdr::Parser> *parsers;
};


MsgHdrParsTab *ReqHdr::TheParsTab = 0;
MsgHdrParsTab *RepHdr::TheParsTab = 0;


/* MsgHdr */

MsgHdr::MsgHdr(const MsgHdrParsTab &aTab): theParsTab(aTab) {
	reset();
}

void MsgHdr::reset() {
	theHdrSize = 0;
	theContSize = -1;
	theDate = Time();
	theXactId.clear();
	theHttpVer = -1;
	theKeepAliveConn = kaDefault;

	theBufBeg = theBufEnd = theSrchPtr = 0;
	theFields.reset();
	theSrchState = ssFirst;

	// do not reset parsing tables
}

bool MsgHdr::keepAlive() const {
	if (theHttpVer >= 1.1) // 1.1: keep unless told otherwise
		return theKeepAliveConn != kaNo;
	else                   // 1.0: keep if explicitly told so
		return theKeepAliveConn == kaYes;
}

// note: buf does not have to be zero-terminated!
bool MsgHdr::parse(const char *buf, Size sz) {
	if (!theBufBeg) { // have not started the search yet
		theBufBeg = theSrchPtr = buf;
		Assert(theSrchState == ssFirst);
	} else {           // continue search
		Assert(theBufBeg == buf);
		Assert(theSrchState != ssFound);
	}

	theBufEnd = buf + sz; // to be refined later
	do {
		// search for LF
		if (theSrchState == ssFirst) {
			do {
				if (*theSrchPtr++ == '\n') {
					theSrchState = ssSkip;
					break;
				}
			} while (theSrchPtr < theBufEnd);
		}

		// LF after skipping optional CRs means and-of-headers
		while (theSrchState == ssSkip && theSrchPtr < theBufEnd) {
			if (*theSrchPtr == '\n') {
				theSrchState = ssFound;
			} else
			if (*theSrchPtr != '\r') {
				theFields.append(theSrchPtr); // start of a header!
				theSrchState = ssFirst;
			}
			++theSrchPtr;
		}
	} while (theSrchPtr < theBufEnd && theSrchState != ssFound);

	if (theSrchState != ssFound)
		return false;

	// found end-of-headers!
	theBufEnd = theSrchPtr;
	theHdrSize = theBufEnd - theBufBeg;

	// now parse known fields
	// luckily, we already know field starts!
	parseFields();
	return true;
}

void MsgHdr::parseFields() {
	const char *eoh = theBufEnd;
	parseRLine(theBufBeg, theFields.count() ? theFields[0] : theBufEnd);

	for (int i = theFields.count()-1; i >= 0; --i) {
		const char *hdr = theFields[i];
		const int len = eoh-hdr; // approximate (includes crlfs)
		const int id = theParsTab.ids->lookup(hdr, len);
		if (id > 0) {
			const char *val = hdr + theParsTab.ids->string(id).len();
			while (isspace(*val)) ++val;
			Parser p = theParsTab.parsers->item(id);
			(this->*p)(val, eoh);
		}
		eoh = hdr;
	}
}

bool MsgHdr::parseUniqId(UniqId &id, const char *buf, const char *) {
	int a = 0, b = 0, c = 0;
	const char *p = 0;
	if (isInt(buf, a, &p, 16) && *p == '.' &&
		isInt(p+1, b, &p, 16) && *p == ':' &&
		isInt(p+1, c, 0, 16))
		return id = UniqId(a, b, c);

	return false;
}

bool MsgHdr::parseContLen(const char *buf, const char *) {
	theContSize = xatoi(buf, -1);
	return theContSize >= 0;
}

bool MsgHdr::parseDate(const char *buf, const char *eoh) {
    theDate = HttpDateParse(buf, eoh - buf);
    return theDate >= 0;
}

bool MsgHdr::parseXXact(const char *buf, const char *eoh) {
	return parseUniqId(theXactId, buf, eoh);
}


/* XXX: Connection and other headers may have a _list_ of options */

bool MsgHdr::parseConnection(const char *buf, const char *) {
	if (!strncasecmp("close", buf, 5))
		theKeepAliveConn = kaNo;
	else
	if (!strncasecmp("keep", buf, 4))
		theKeepAliveConn = kaYes;
	else
		return false;
	return true;
}

// adds definitions common to replies and requests
void MsgHdr::Configure(MsgHdrParsTab &tab) {
	AddParser("Date:", &MsgHdr::parseDate, tab);
	AddParser("Content-Length:", &MsgHdr::parseContLen, tab);
	AddParser("Connection:", &MsgHdr::parseConnection, tab);
	AddParser("Proxy-Connection:", &MsgHdr::parseConnection, tab);
	AddParser("X-Xact:", &MsgHdr::parseXXact, tab);
}

int MsgHdr::AddParser(const String &field, Parser parser, MsgHdrParsTab &where) {
	const int id = where.ids->add(field);
	where.parsers->put(parser, id);
	return id;
}

void MsgHdr::store(OLog &log) const {
	log
		<< theHdrSize
		<< (int)theDate.sec()
		<< theContSize
		<< theXactId
		<< theHttpVer
		<< theKeepAliveConn
		;
}

// these should never be called
inline bool dontCallMe() { Assert(0); return false; }
bool MsgHdr::parsePragma(const char *, const char *)  { return dontCallMe(); }
bool MsgHdr::parseCControl(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseHost(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseLMT(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseExpires(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseIMS(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseXOidShortage(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseXFailedOid(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseXCltWss(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseXSrvWid(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseXSrvSets(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseXSrvOids(const char *, const char *) { return dontCallMe(); }


/* ReqHdr */

ReqHdr::ReqHdr(): MsgHdr(*TheParsTab),
	theOidShortage(0), theFailedOid(-1), theCltWss(-1) {
}

void ReqHdr::reset() {
	MsgHdr::reset();
	theIMS = Time();
	theOidRec.reset();
	theOidShortage = 0;
	theFailedOid = -1;
	theCltWss = -1;
}

bool ReqHdr::parseRLine(const char *buf, const char *eorl) {
	if (strncasecmp("GET ", buf, 4) == 0) {
		buf += 4;
		parseUrl(buf, eorl);
		if (const char *ver = StrBoundChr(buf, ' ', eorl))
			isDbl(ver + 6, theHttpVer);
		return true;
	}
	return false;
}

bool ReqHdr::parseUrl(const char *&buf, const char *end) {
	// see if there is a protocol://host prefix
	if (*buf != '/')
		parseHostInURL(buf, end);
	if (const char *wid = StrBoundChr(buf, 'w', end))
		parseUniqId(theOidRec.wid, buf = wid+1, end);
	if (const char *tid = StrBoundChr(buf, 't', end))
		isInt(tid + 1, theOidRec.tid, &buf, 16);
	if (const char *oid = StrBoundChr(buf, '_', end))
		isInt(oid + 1, theOidRec.oid, &buf, 16);
	return true;
}

bool ReqHdr::parseHostInURL(const char *&start, const char *eorl) {
	const char *buf = start;
	if (const char *n = StrBoundChr(buf, ':', eorl)) {
		if (*++n == '/' && *++n == '/') {
			buf = n+1;

			const char *end = StrBoundChr(buf, '/', eorl);
			if (!end)
				end = eorl;

			if (parseHost(buf, end)) {
				start = end;
				return true;
			}
		}
	}
	return false;
}

bool ReqHdr::parseHost(const char *buf, const char *eoh) {
	int port = 80; // XXX: default should depend on protocol
	if (const char *p = StrBoundChr(buf, ':', eoh)) {
		if (isInt(p + 1, port))
			eoh = p;
		else 
			return false; // malformed port number
	}
				
	// convert a.b.c.d to an int address
	// this is a poor man's inet_aton()...
	int parts[4];
	if (!isInt(buf, parts[3], &buf))
		return false;
	for (int i = 2; i >= 0 && buf < eoh; --i) {
		if (*buf != '.')
			return false;
		if (!isInt(buf+1, parts[i], &buf))
			return false;
		if (parts[i] < 0 || parts[i] > 255)
			return false;
	}

	if (buf != eoh)
		return false; // trailing garbage

	in_addr name;
	name.s_addr =
		(parts[0] << 24) | (parts[1] << 16) | (parts[2] << 8) | parts[3];
	theHost = NetAddr(name, port);
	return true;
}

bool ReqHdr::parseIMS(const char *buf, const char *eoh) {
	theIMS = HttpDateParse(buf, eoh - buf);
	return theIMS >= 0;
}

bool ReqHdr::parseXOidShortage(const char *buf, const char *) {
	theOidShortage = xatoi(buf, 0);
	return true;
}

bool ReqHdr::parseXFailedOid(const char *buf, const char *) {
	theFailedOid = xatoi(buf, 0);
	return true;
}

bool ReqHdr::parseXCltWss(const char *buf, const char *) {
	theCltWss = xatoi(buf, 0);
	return true;
}

void ReqHdr::store(OLog &log) const {
	MsgHdr::store(log);
	log
		<< theHost
		<< (int)theIMS.sec()
		<< theCltWss
		;
}

void ReqHdr::Configure() {
	TheParsTab = new MsgHdrParsTab();
	MsgHdr::Configure(*TheParsTab);
	AddParser("Host:", &MsgHdr::parseHost, *TheParsTab);
	AddParser("If-Modified-Since:", &MsgHdr::parseIMS, *TheParsTab);
	AddParser("X-Oid-Shortage:", &MsgHdr::parseXOidShortage, *TheParsTab);
	AddParser("X-Failed-Oid:", &MsgHdr::parseXFailedOid, *TheParsTab);
	AddParser("X-Clt-Wss:", &MsgHdr::parseXCltWss, *TheParsTab);
}

void ReqHdr::Clean() {
	delete TheParsTab;
	TheParsTab = 0;
}


/* RepHdr */

RepHdr::RepHdr(): MsgHdr(*TheParsTab),
	theStatus(-1),
	theSecOid(-1), theNewOid(-1), theNewOidCnt(0),
	theSrvWss(-1), theHotSetPos(-1),
	isCachable(true) {
}

void RepHdr::reset() {
	MsgHdr::reset();
	theStatus = -1;
	theLMT = theExpires = Time();
	theSrvWid.clear();
	theSecOid = theNewOid = -1;
	theNewOidCnt = 0;
	theSrvWss = theHotSetPos = -1;
	isCachable = true;
}

/* RFC 2616: All responses to the HEAD request method MUST NOT include
 * a message-body. All 1xx , 204, and 304 responses MUST NOT include a
 * message-body. All other responses do include a message-body.
 */
bool RepHdr::expectBody() const {
	// note: we cannot handle the HEAD case here; the caller should
	if ((100 <= theStatus && theStatus < 200) ||
		theStatus == scNoContent ||
		theStatus == scNotModified)
		return false;
	return true;
}

Time RepHdr::calcLmt() const {
	if (theLMT >= 0)
		return theLMT;
	if (theDate >= 0)
		return theDate;
	return TheClock;
}

bool RepHdr::parseRLine(const char *buf, const char *) {
	// what is the actual version format in the specs??
	if (strncasecmp("HTTP/", buf, 5) != 0)
		return false;
	if (!isDbl(buf+5, theHttpVer, &buf))
		return false;
	return isInt(buf+1, theStatus);
}

bool RepHdr::parseLMT(const char *buf, const char *eoh) {
    theLMT = HttpDateParse(buf, eoh - buf);
    return theLMT >= 0;
}

bool RepHdr::parseExpires(const char *buf, const char *eoh) {
    theExpires = HttpDateParse(buf, eoh - buf);
    return theExpires >= 0;
}

bool RepHdr::parsePragma(const char *buf, const char *) {
	if (!strncasecmp("no-cache", buf, 8))
		isCachable = false;
	else
		return false;
	return true;
}

bool RepHdr::parseCControl(const char *buf, const char *) {
	if (!strncasecmp("no-cache", buf, 8)) {
		isCachable = false;
		return true;
	}
	return false;
}

bool RepHdr::parseXSrvWid(const char *buf, const char *eoh) {
	return parseUniqId(theSrvWid, buf, eoh);
}

bool RepHdr::parseXSrvSets(const char *buf, const char *) {
	const char *p = 0;
	if (isInt(buf, theSrvWss, &p) && *p == '.') {
		theHotSetPos = xatoi(p+1);
		return true;
	}
	return false;
}

bool RepHdr::parseXSrvOids(const char *buf, const char *) {
	const char *p = 0;
	if (isInt(buf, theSecOid, &p) && *p == ':' &&
		isInt(p+1, theNewOid, &p) && *p == '.') {
		theNewOidCnt = xatoi(p+1);
		return true;
	}
	return false;
}

void RepHdr::store(OLog &log) const {
	MsgHdr::store(log);
	log
		<< theStatus
		<< (int)theLMT.sec()
		<< (int)theExpires.sec()
		<< theSrvWid
		<< theSecOid
		<< theSrvWss
		<< theHotSetPos
		<< isCachable
		;
}

void RepHdr::Configure() {
	TheParsTab = new MsgHdrParsTab();
	MsgHdr::Configure(*TheParsTab);
	AddParser("Last-Modified:", &MsgHdr::parseLMT, *TheParsTab);
	AddParser("Expires:", &MsgHdr::parseExpires, *TheParsTab);
	AddParser("Pragma:", &MsgHdr::parsePragma, *TheParsTab);
	AddParser("Cache-Control:", &MsgHdr::parseCControl, *TheParsTab);
	AddParser("X-Srv-Wid:", &MsgHdr::parseXSrvWid, *TheParsTab);
	AddParser("X-Srv-Sets:", &MsgHdr::parseXSrvSets, *TheParsTab);
	AddParser("X-Srv-Oids:", &MsgHdr::parseXSrvOids, *TheParsTab);
}

void RepHdr::Clean() {
	delete TheParsTab;
	TheParsTab = 0;
}


/* MsgHdrParsTab */

MsgHdrParsTab::MsgHdrParsTab() {
	ids = new StrIdentifier;
	parsers = new Array<MsgHdr::Parser>;
}

MsgHdrParsTab::~MsgHdrParsTab() {
	delete ids; ids = 0;
	delete parsers; parsers = 0;
}



/* initialization */

int HttpHdrsInit::TheUseCount = 0;

void HttpHdrsInit::init() {
	ReqHdr::Configure();
	RepHdr::Configure();
}

void HttpHdrsInit::clean() {
	RepHdr::Clean();
	ReqHdr::Clean();
}
