
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "polygraph.h"

#include "xstd/Clock.h"
#include "xstd/gadgets.h"
#include "runtime/ErrorMgr.h"
#include "runtime/LogComment.h"
#include "runtime/SharedOpts.h"
#include "runtime/polyErrors.h"
#include "server/SrvHistory.h"


static const int TheNewOidsPerRepMax = 16;
static const Size TheAdvWinSizeMax = Size::KB(100);

SrvHistory::SrvHistory(): theSecOid(0), theAdvOid(0), thePendCnt(0),
	theWss(-1), theSecTime(TheClock) {
}

void SrvHistory::window(int capacity) {
	theAdvWin.resize(capacity);
}

int SrvHistory::newOid() {
	// if the gap is getting too big
	if (theAdvOid - theSecOid >= theAdvWin.capacity()) {
		const bool reportReclaim = ReportError(errUnclaimOid);

		const int stuckOid = theSecOid + 1;
		bool reclaim = true;
		if (thePendCnt > theAdvWin.capacity()/2) {
			if (2*theAdvWin.size() <= TheAdvWinSizeMax) {
				// we probably cannot keep up with the client side
				theAdvWin.grow(Max(1, 2*theAdvWin.capacity()), stuckOid);
				if (ReportError(errNidReqFlood))
					Comment(4) << "grew server oid map to " << theAdvWin.size() << endc;
				reclaim = false;
			} else {
				// we are prohibited to grow any larger
				ReportError(errNidMapLimit);
				reclaim = true;
			}
		}

		if (reclaim) {
			const Time delay = TheClock - theSecTime;
			if (reportReclaim && delay >= Time::Sec(5))
				Comment(4) << "reclaiming after at least " << delay << endc;

			// assume that oid got stuck due to an error and "reclaim" it
			Assert(!requested(stuckOid));
			noteOid(stuckOid);
		}

		// check that we fixed the problem
		Assert(theAdvOid - theSecOid < theAdvWin.capacity());
	}

	++theAdvOid;
	requested(theAdvOid, false);
	return theAdvOid;
}

void SrvHistory::noteOid(int oid) {
	if (oid <= theSecOid || requested(oid)) // nothing new
		return;

	if (oid > theAdvOid) {
		ReportError(errUnsolicOid);
		return;
	}
	
	requested(oid, true);

	// increment secured level if possible
	if (oid == theSecOid + 1) {
		do {
			++theSecOid;
		} while (theSecOid < theAdvOid && requested(theSecOid+1));
		theSecTime = TheClock;
	}
}

void SrvHistory::noteWss(int wss) {
	if (wss > theWss)
		theWss = wss;
}

int SrvHistory::newOids(int needCnt, int &resCnt) {
	resCnt = 0;
	
	int firstOid = -1;
	while (needCnt > 0 && resCnt < TheOpts.theNidsPerMsgMax) {
		const int oid = newOid();
		if (firstOid < 0)
			firstOid = oid;
		else
			Assert(firstOid + resCnt == oid);
		needCnt--;
		resCnt++;
	}

	return firstOid;
}

void SrvHistory::store(OLog &log) const {
	log
		// << theAdvWin // not storing bit mask for now
		<< theSecOid
		<< theAdvOid
		// << thePendCnt // not storing for now
		<< theWss
		;
}
