
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "xstd.h"

#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_SOCKIO_H
#include <sys/sockio.h>  /* for SIOCGIFCONF on Solaris */
#endif
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netinet/tcp.h> /* for TCP_NODELAY */
#include <net/if.h>      /* for ifconf */
#include <arpa/inet.h>
#include <netdb.h>

#include "Socket.h"
#include "Assert.h"

bool Socket::IsEnabled = true;
int  Socket::TheMaxLevel = -1;
int  Socket::TheLevel = 0;

// called when theFd is set to a new value
bool Socket::noteNewFD() {
	if (theFD < 0)
		return false;

	TheLevel++;

	// most MaxLevel users assume that MaxLevel means max FD!
	if (TheMaxLevel >= 0 && theFD >= TheMaxLevel) {
		close();
		errno = EMFILE;
		return false;
	}

	return true;
}

bool Socket::create() {
	return create(PF_INET, SOCK_STREAM, 0);
}

bool Socket::create(int domain, int type, int protocol) {

	if (TheMaxLevel >= 0 && TheLevel >= TheMaxLevel) {
		errno = EMFILE;
		return false;
	}
	
	theFD = ::socket(domain, type, protocol);
	return noteNewFD();
}

bool Socket::bind(const NetAddr &addr) {
	static sockaddr_in S;

	S.sin_family = AF_INET;
	S.sin_port = htons(addr.port());
	if (addr.addrA())
		S.sin_addr = addr.addrN();
	else
		S.sin_addr.s_addr = htonl(INADDR_ANY);

	return !IsEnabled ||
		::bind(theFD, (sockaddr *) &S, sizeof(S)) == 0;
}

bool Socket::connect(const NetAddr &addr) {
	static sockaddr_in s;

	s.sin_family = AF_INET;
	s.sin_addr = addr.addrN();
	s.sin_port = htons(addr.port());

	if (IsEnabled && ::connect(theFD, (sockaddr *) &s, sizeof(s)) < 0)
		return errno == EINPROGRESS;
	return true;
}

bool Socket::listen() {
	return !IsEnabled ||
		::listen(theFD, FD_SETSIZE) == 0;
}

Socket Socket::accept(sockaddr *addr_p, socklen_t *addr_len_p) {

	if (TheMaxLevel >= 0 && TheLevel >= TheMaxLevel) {
		errno = EMFILE;
		return -1;
	}

	sockaddr addr;
	socklen_t addr_len = sizeof(addr);
	if (!addr_p) {
		addr_p = &addr;
		addr_len_p = &addr_len;
	} else
		Assert(addr_len_p);

	const int fd = IsEnabled ?
		::accept(theFD, addr_p, addr_len_p) :
		(errno = EWOULDBLOCK, -1);

	Socket s(fd);
	s.noteNewFD();
	return s;
}

Size Socket::read(void *buf, Size sz) {
	return IsEnabled && sz > 0 ? 
		(Size)::read(theFD, buf, sz) :
		(Size)0;
}

Size Socket::write(const void *buf, Size sz) {
	return IsEnabled && sz > 0 ? 
		(Size)::write(theFD, buf, sz) :
		sz;
}

Size Socket::recvFrom(void *buf, Size sz, NetAddr &addr, int flags) {
	if (!IsEnabled || sz <= 0)
		return 0;

	sockaddr_in sa;
	sa.sin_family = AF_INET;
	sa.sin_addr = addr.addrN();
	sa.sin_port = htons(addr.port());

	socklen_t addr_len = sizeof(sa);

	const Size readSz =
		(Size)::recvfrom(theFD, buf, sz, flags, (sockaddr *) &sa, &addr_len);

	if (readSz >= 0) {
		addr.addr(sa.sin_addr);
		addr.port(ntohs(sa.sin_port));
	}
	return readSz;
}

Size Socket::sendTo(const void *buf, Size sz, const NetAddr &addr, int flags) {
	sockaddr_in sa;
	sa.sin_family = AF_INET;
	sa.sin_addr = addr.addrN();
	sa.sin_port = htons(addr.port());

	return IsEnabled && sz > 0 ?
		(Size)::sendto(theFD, buf, sz, flags, (const sockaddr *) &sa, sizeof(sa)) :
		(Size)0;
}

bool Socket::close() {
	if (theFD >= 0) {
		const int res = ::close(theFD);
		theFD = -1;
		TheLevel--;
		return res == 0;
	}
	errno = EBADF;
	return false;
}

bool Socket::sendBuf(Size sz) {
#ifdef SO_SNDBUF
	return setOpt(SO_SNDBUF, (const void*)&sz, sizeof(sz));
#else
	return unsupOpt(&sz);
#endif
}

bool Socket::recvBuf(Size sz) {
#ifdef SO_RCVBUF
	return setOpt(SO_RCVBUF, (const void*)&sz, sizeof(sz));
#else
	return unsupOpt(&sz);
#endif
}

bool Socket::getIfConf(struct ifconf &cfg) {
#ifdef SIOCGIFCONF
	return ioctl(theFD, SIOCGIFCONF, &cfg) >= 0;
#else
	errno = EINVAL;
	return false;
#endif
}

bool Socket::getIfAddr(const char *ifname, struct sockaddr &addr) {
#ifdef SIOCGIFADDR
	struct ifreq r;
	strncpy(r.ifr_name, ifname, sizeof(r.ifr_name));
	if (ioctl(theFD, SIOCGIFADDR, &r) < 0)
		return false;
	addr = r.ifr_addr;
	return true;
#else
	errno = EINVAL;
	return false;
#endif
}


bool Socket::getIfBcastAddr(const char *ifname, struct sockaddr &addr) {
#ifdef SIOCGIFADDR
	struct ifreq r;
	strncpy(r.ifr_name, ifname, sizeof(r.ifr_name));
	if (ioctl(theFD, SIOCGIFBRDADDR, &r) < 0)
		return false;
	addr = r.ifr_addr;
	return true;
#else
	errno = EINVAL;
	return false;
#endif
}


bool Socket::addIfAddr(const ifaliasreq &req) {
#ifdef SIOCAIFADDR
	return ioctl(theFD, SIOCAIFADDR, &req) >= 0;
#else
	errno = EINVAL;
	return false;
#endif
}

bool Socket::delIfAddr(const ifaliasreq &req) {
#ifdef SIOCDIFADDR
	return ioctl(theFD, SIOCDIFADDR, &req) >= 0;
#else
	errno = EINVAL;
	return false;
#endif
}

bool Socket::configure(const SockOpt &opt) {
	if (opt.nagle && !tcpDelay(opt.nagle > 0))
		return false;

	if (opt.linger >= 0 && !linger(opt.linger))
		return false;

	return true;
}

int Socket::flags() const {
	return fcntl(theFD, F_GETFL);
}

/* local port */
int Socket::lport() const {
	sockaddr_in S;
	socklen_t len = sizeof(S);

	if (getsockname(theFD, (sockaddr*)&S, &len) == 0)
		return ntohs(S.sin_port);
	else
		return -1;
}

/* remote (peer) port */
int Socket::rport() const {
	sockaddr_in S;
	socklen_t len = sizeof(S);

	if (getpeername(theFD, (sockaddr*)&S, &len) == 0)
		return ntohs(S.sin_port);
	else
		return -1;
}

Time Socket::rcvTimeout() const {
#ifdef SO_RCVTIMEO
	Time t;
	socklen_t len = sizeof(t);
	if (getOpt(SO_RCVTIMEO, (void*)&t, &len))
		return t;
#endif
	return Time();
}

Size Socket::sendBuf() const {
#ifdef SO_SNDBUF
	Size sz;
	socklen_t len = sizeof(sz);
	if (getOpt(SO_SNDBUF, (void*)&sz, &len))
		return sz;
#endif
	return Size();
}

Size Socket::recvBuf() const {
#ifdef SO_RCVBUF
	Size sz;
	socklen_t len = sizeof(sz);
	if (getOpt(SO_RCVBUF, (void*)&sz, &len))
		return sz;
#endif
	return Size();
}

/*
 * Note: Solaris 2.4's socket emulation doesn't allow you
 * to determine the error from a failed non-blocking
 * connect and just returns EPIPE. -- fenner@parc.xerox.com
 */
int Socket::errNo() const {
	int e;
	socklen_t len = sizeof(e);

	if (getOpt(SO_ERROR, (void*)&e, &len))
		return errno = e;
	else
		return -1;
}

bool Socket::peerAddr(sockaddr *S, socklen_t &len) const {
	return getpeername(theFD, S, &len) == 0;
}

NetAddr Socket::peerAddr() const {
	sockaddr_in s;
	socklen_t len = sizeof(s);
	if (peerAddr((sockaddr *) &s, len))
		return NetAddr(inet_ntoa(s.sin_addr), ntohs(s.sin_port));
	else
		return NetAddr();
}

bool Socket::reuseAddr(bool set) {
#ifdef SO_REUSEADDR
	return setOpt(SO_REUSEADDR, set);
#else
	return unsupOpt(&set);
#endif
}

bool Socket::linger(Time delay) {
#ifdef SO_LINGER
	static struct linger l = { delay.sec() > 0, delay.sec() };
	return setOpt(SO_LINGER, &l, sizeof(l));
#else
	return unsupOpt(&delay);
#endif
}

bool Socket::blocking(bool block) {

	const int oldFlags = flags();
	int newFlags = oldFlags;

	if (block)
		newFlags &= ~O_NONBLOCK;
	else
		newFlags |=  O_NONBLOCK;

	if (newFlags == oldFlags)
		return true;

	return fcntl(theFD, F_SETFL, newFlags) != -1;
}

bool Socket::reusePort(bool set) {
#ifdef SO_REUSEPORT
	return setOpt(SO_REUSEPORT, set);
#else
	return unsupOpt(&set);
#endif
}

bool Socket::tcpDelay(bool set) {
#ifdef TCP_NODELAY
	int optVal = (int)!set;
	// must cast to char* on Solaris
	return ::setsockopt(theFD, IPPROTO_TCP, TCP_NODELAY, (const char*)&optVal, sizeof(optVal)) == 0;
#else
	return unsupOpt(&set);
#endif
}

bool Socket::isBlocking() const {
	return flags() & O_NONBLOCK == 0;
}


bool Socket::setOpt(int optName, bool set) {
	int optVal = (int)set;
	return setOpt(optName, &optVal, sizeof(optVal));
}

bool Socket::setOpt(int optName, const void *optVal, socklen_t optLen) {
	return setOpt(SOL_SOCKET, optName, (const char*)optVal, optLen);
}

bool Socket::getOpt(int optName, void *optVal, socklen_t *optLen) const {
	return getOpt(SOL_SOCKET, optName, (char*)optVal, optLen);
}

bool Socket::setOpt(int level, int optName, const void *optVal, socklen_t optLen) {
	// must cast to char* on Solaris
	return ::setsockopt(theFD, level, optName, (const char*)optVal, optLen) == 0;
}

bool Socket::getOpt(int level, int optName, void *optVal, socklen_t *optLen) const {
	// must cast to char* on Solaris
	return ::getsockopt(theFD, level, optName, (char*)optVal, optLen) == 0;
}

// the argument is used by the caller to prevent "arg not used" warnings
bool Socket::unsupOpt(void *) const {
	errno = ENOPROTOOPT;
	return false;
}
